""":mod:`itunesiap.shortcut`"""
from .request import Request


def verify(
        receipt_data, password=None, exclude_old_transactions=False, **kwargs):
    """Shortcut API for :class:`itunesiap.request.Request`.

    See also:
        - Receipt_Validation_Programming_Guide_.

    .. _Receipt_Validation_Programming_Guide: https://developer.apple.com/library/content/releasenotes/General/ValidateAppStoreReceipt/Chapters/ValidateRemotely.html

    :param str receipt_data: :class:`itunesiap.request.Request` argument.
        An iTunes receipt data as Base64 encoded string.
    :param str password: :class:`itunesiap.request.Request` argument. Optional.
        Only used for receipts that contain auto-renewable subscriptions. Your
        app's shared secret (a hexadecimal string).
    :param bool exclude_old_transactions: :class:`itunesiap.request.Request`
        argument. Optional. Only used for iOS7 style app receipts that contain
        auto-renewable or non-renewing subscriptions. If value is true,
        response includes only the latest renewal transaction for any
        subscriptions.

    :param itunesiap.environment.Environment env: Set base environment value.
        See :mod:`itunesiap.environment` for detail.
    :param float timeout: :func:`itunesiap.request.Request.verify` argument.
        Keyword-only optional. The value is connection timeout of the verifying
        request. The default value is 30.0 when no `env` is given.
    :param bool use_production: :func:`itunesiap.request.Request.verify`
        argument. Keyword-only optional. The value is weather verifying in
        production server or not. The default value is :class:`bool` True
        when no `env` is given.
    :param bool use_sandbox: :func:`itunesiap.request.Request.verify`
        argument. Keyword-only optional. The value is weather verifying in
        sandbox server or not. The default value is :class:`bool` False
        when no `env` is given.

    :param bool verify_ssl: :func:`itunesiap.request.Request.verify` argument.
        Keyword-only optional. The value is weather enabling SSL verification
        or not. WARNING: DO NOT TURN IT OFF WITHOUT A PROPER REASON. IF YOU
        DON'T UNDERSTAND WHAT IT MEANS, NEVER SET IT YOURSELF.
    :param str proxy_url: Keyword-only optional. A proxy url to access the
        iTunes validation url.

    :return: :class:`itunesiap.receipt.Receipt` object if succeed.
    :raises: Otherwise raise a request exception in :mod:`itunesiap.exceptions`.
    """
    proxy_url = kwargs.pop('proxy_url', None)
    request = Request(
        receipt_data, password, exclude_old_transactions, proxy_url=proxy_url)
    return request.verify(**kwargs)


def aioverify(
        receipt_data, password=None, exclude_old_transactions=False, **kwargs):
    """Shortcut API for :class:`itunesiap.request.Request`.

    Note that python3.4 support is only available at itunesiap==2.5.1

    For params and returns, see :func:`itunesiap.verify`.
    """
    proxy_url = kwargs.pop('proxy_url', None)
    request = Request(
        receipt_data, password, exclude_old_transactions, proxy_url=proxy_url)
    return request.aioverify(**kwargs)



""":mod:`itunesiap.request`"""

from itunesiap.verify_requests import RequestsVerify

try:
    from itunesiap.verify_aiohttp import AiohttpVerify
except (SyntaxError, ImportError, AttributeError):  # pragma: no cover
    class AiohttpVerify(object):
        pass


class RequestBase(object):

    PRODUCTION_VALIDATION_URL = "https://buy.itunes.apple.com/verifyReceipt"
    SANDBOX_VALIDATION_URL = "https://sandbox.itunes.apple.com/verifyReceipt"
    STATUS_SANDBOX_RECEIPT_ERROR = 21007

    def __init__(
            self, receipt_data, password=None, exclude_old_transactions=False,
            **kwargs):
        self.receipt_data = receipt_data
        self.password = password
        self.exclude_old_transactions = exclude_old_transactions
        self.proxy_url = kwargs.pop('proxy_url', None)
        if kwargs:  # pragma: no cover
            raise TypeError(
                u"__init__ got unexpected keyword argument {}".format(
                    ', '.join(kwargs.keys())))

    def __repr__(self):
        return u'<Request({0}...)>'.format(self.receipt_data[:20])

    @property
    def request_content(self):
        """Instantly built request body for iTunes."""
        request_content = {
            'receipt-data': self.receipt_data,
            'exclude-old-transactions': self.exclude_old_transactions}
        if self.password is not None:
            request_content['password'] = self.password
        return request_content


class Request(RequestBase, RequestsVerify, AiohttpVerify):
    """Validation request with raw receipt.

    Use `verify` method to try verification and get Receipt or exception.
    For detail, see also the Apple document: `<https://developer.apple.com/library/content/releasenotes/General/ValidateAppStoreReceipt/Chapters/ValidateRemotely.html>`_.

    :param str receipt_data: An iTunes receipt data as Base64 encoded string.
    :param str password: Only used for receipts that contain auto-renewable subscriptions. Your app's shared secret (a hexadecimal string).
    :param bool exclude_old_transactions: Only used for iOS7 style app receipts that contain auto-renewable or non-renewing subscriptions. If value is true, response includes only the latest renewal transaction for any subscriptions.
    :param proxy_url: A proxy url to access the iTunes validation url.
        (It is an attribute of :func:`verify` but misplaced here)
    """



""":mod:`itunesiap.environment`

:class:`Environment` is designed to pass pre-defined policies in easy way.
The major use cases are provided as pre-defined constants.

How to use environments
-----------------------

The default policy is `default` and it is the same to `production`. When your
development lifecycle is proceed, you want to change it to `sandbox` or
`review`.

The recommended way to use environments is passing the value to
:func:`itunesiap.verify` function as keyword argument `env`.

.. sourcecode:: python

    >>> itunesiap.verify(receipt, env=itunesiap.env.production)
    >>> itunesiap.verify(receipt, env=itunesiap.env.sandbox)
    >>> itunesiap.verify(receipt, env=itunesiap.env.review)


Review mode
-----------

This is useful when your server is being used both for real users and Apple
reviewers.
Using review mode for a real service is possible, but be aware of: it is not
100% safe. Your testers can getting advantage of free IAP in production
version.
A rough solution what I suggest is:

.. sourcecode:: python

    >>> if client_version == review_version:
    >>>     env = itunesiap.env.review
    >>> else:
    >>>     env = itunesiap.env.production
    >>>
    >>> itunesiap.verify(receipt, env=env)


Environment
-----------
"""
from itunesiap.tools import deprecated

__all__ = ('Environment', 'default', 'production', 'sandbox', 'review')


class EnvironmentStack(list):
    @deprecated
    def push(self, env):
        self.append(env)


class Environment(object):
    """Environment provides option preset for `Request`. `default` is default.

    By passing an environment object to :func:`itunesiap.verify` or
    :func:`itunesiap.request.Request.verify` function, it replaces verifying
    policies.
    """

    ITEMS = (
        'use_production', 'use_sandbox', 'timeout', 'exclude_old_transactions',
        'verify_ssl')

    def __init__(self, **kwargs):
        self.use_production = kwargs.get('use_production', True)
        self.use_sandbox = kwargs.get('use_sandbox', False)
        self.timeout = kwargs.get('timeout', None)
        self.exclude_old_transactions = kwargs.get('exclude_old_transactions', False)
        self.verify_ssl = kwargs.get('verify_ssl', True)

    def __repr__(self):
        return u'<{self.__class__.__name__} use_production={self.use_production} use_sandbox={self.use_sandbox} timeout={self.timeout} exclude_old_transactions={self.exclude_old_transactions} verify_ssl={self.verify_ssl}>'.format(self=self)

    def clone(self, **kwargs):
        """Clone the environment with additional parameter override"""
        options = self.extract()
        options.update(**kwargs)
        return self.__class__(**options)

    def override(self, **kwargs):
        """Override options in kwargs to given object `self`."""
        for item in self.ITEMS:
            if item in kwargs:
                setattr(self, item, kwargs[item])

    def extract(self):
        """Extract options from `self` and merge to `kwargs` then return a new
        dictionary with the values.
        """
        options = {}
        for item in self.ITEMS:
            options[item] = getattr(self, item)
        return options

    # backward compatibility features
    _stack = EnvironmentStack()

    @deprecated
    def push(self):  # pragma: no cover
        self._stack.push(self)

    @deprecated
    def __enter__(self):  # pragma: no cover
        self._ctx_id = len(self._stack)
        self._stack.push(self)
        return self

    @deprecated
    def __exit__(self, exc_type, exc_value, tb):  # pragma: no cover
        self._stack.pop(self._ctx_id)

    @classmethod
    @deprecated
    def current(cls):  # pragma: no cover
        return cls._stack[-1]


#: Use only production server with 30 seconds of timeout.
default = Environment(use_production=True, use_sandbox=False, timeout=30.0, verify_ssl=True)
#: Use only production server with 30 seconds of timeout.
production = Environment(use_production=True, use_sandbox=False, timeout=30.0, verify_ssl=True)
#: Use only sandbox server with 30 seconds of timeout.
sandbox = Environment(use_production=False, use_sandbox=True, timeout=30.0, verify_ssl=True)

review = Environment(use_production=True, use_sandbox=True, timeout=30.0, verify_ssl=True)
#: Use both production and sandbox servers with 30 seconds of timeout.

unsafe = Environment(use_production=True, use_sandbox=True, verify_ssl=False)


Environment._stack.append(default)  # for backward compatibility


@deprecated
def current():
    return Environment.current()




import json
import functools
import requests

from . import receipt
from . import exceptions
from .environment import Environment


class InvalidReceiptResponse(exceptions.InvalidReceipt, receipt.Response):
    pass


class RequestsVerify(object):
    def verify_from(self, url, timeout=None, verify_ssl=True):
        """The actual implemention of verification request.

        :func:`verify` calls this method to try to verifying for each servers.

        :param str url: iTunes verification API URL.
        :param float timeout: The value is connection timeout of the verifying
            request. The default value is 30.0 when no `env` is given.
        :param bool verify_ssl: SSL verification.

        :return: :class:`itunesiap.receipt.Receipt` object if succeed.
        :raises: Otherwise raise a request exception.
        """
        post_body = json.dumps(self.request_content)
        requests_post = requests.post
        if self.proxy_url:
            protocol = self.proxy_url.split('://')[0]
            requests_post = functools.partial(requests_post, proxies={protocol: self.proxy_url})
        if timeout is not None:
            requests_post = functools.partial(requests_post, timeout=timeout)
        try:
            http_response = requests_post(url, post_body, verify=verify_ssl)
        except requests.exceptions.RequestException as e:
            raise exceptions.ItunesServerNotReachable(exc=e)

        if http_response.status_code != 200:
            raise exceptions.ItunesServerNotAvailable(http_response.status_code, http_response.content)

        response_data = json.loads(http_response.content.decode('utf-8'))
        response = receipt.Response(response_data)
        if response.status != 0:
            raise exceptions.InvalidReceipt(response_data=response_data)
        return response

    def verify(self, **options):
        """Try verification with current environment.

        See also:
            - Receipt_Validation_Programming_Guide_.

        .. _Receipt_Validation_Programming_Guide: https://developer.apple.com/library/content/releasenotes/General/ValidateAppStoreReceipt/Chapters/ValidateRemotely.html

        :param itunesiap.environment.Environment env: Override the environment.
        :param float timeout: The value is connection timeout of the verifying
            request. The default value is 30.0 when no `env` is given.
        :param bool use_production: The value is weather verifying in
            production server or not. The default value is :class:`bool` True
            when no `env` is given.
        :param bool use_sandbox: The value is weather verifying in
            sandbox server or not. The default value is :class:`bool` False
            when no `env` is given.

        :param bool verify_ssl: The value is weather enabling SSL verification
            or not. WARNING: DO NOT TURN IT OFF WITHOUT A PROPER REASON. IF YOU
            DON'T UNDERSTAND WHAT IT MEANS, NEVER SET IT YOURSELF.

        :return: :class:`itunesiap.receipt.Receipt` object if succeed.
        :raises: Otherwise raise a request exception.
        """
        env = options.get('env')
        if not env:  # backward compitibility
            env = Environment._stack[-1]
        use_production = options.get('use_production', env.use_production)
        use_sandbox = options.get('use_sandbox', env.use_sandbox)
        verify_ssl = options.get('verify_ssl', env.verify_ssl)
        timeout = options.get('timeout', env.timeout)
        assert(env.use_production or env.use_sandbox)

        response = None
        if use_production:
            try:
                response = self.verify_from(self.PRODUCTION_VALIDATION_URL, timeout=timeout, verify_ssl=verify_ssl)
            except exceptions.InvalidReceipt as e:
                if not use_sandbox or e.status != self.STATUS_SANDBOX_RECEIPT_ERROR:
                    raise

        if not response and use_sandbox:
            try:
                response = self.verify_from(self.SANDBOX_VALIDATION_URL, timeout=timeout, verify_ssl=verify_ssl)
            except exceptions.InvalidReceipt:
                raise

        return response



"""
itunes-iap
~~~~~~~~~~

Itunes In-app Purchase verification api.

:copyright: (c) 2013 Jeong YunWon
:license: 2-clause BSD.
"""

from .request import Request
from .receipt import Response, Receipt, InApp
from .shortcut import verify, aioverify

from . import exceptions
from . import environment

exc = exceptions
env = environment  # env.default, env.sandbox, env.review


__version__ = '2.6.1'
__all__ = (
    '__version__', 'Request', 'Response', 'Receipt', 'InApp',
    'verify', 'aioverify',
    'exceptions', 'exc', 'environment', 'env')



import asyncio
import json
import aiohttp

from . import receipt
from . import exceptions
from .environment import default as default_env


class AiohttpVerify:

    async def aioverify_from(self, url, timeout):
        body = json.dumps(self.request_content).encode()
        async with aiohttp.ClientSession() as session:
            try:
                http_response = await session.post(url, data=body, timeout=timeout)
            except asyncio.TimeoutError as e:
                raise exceptions.ItunesServerNotReachable(exc=e)
            if http_response.status != 200:
                response_text = await http_response.text()
                raise exceptions.ItunesServerNotAvailable(http_response.status, response_text)
            response_body = await http_response.text()
            response_data = json.loads(response_body)
            response = receipt.Response(response_data)
            if response.status != 0:
                raise exceptions.InvalidReceipt(response_data)
            return response

    async def aioverify(self, **options):
        """Try to verify the given receipt with current environment.

        Note that python3.4 support is only available at itunesiap==2.5.1

        See also:
            - Receipt_Validation_Programming_Guide_.

        .. _Receipt_Validation_Programming_Guide: https://developer.apple.com/library/content/releasenotes/General/ValidateAppStoreReceipt/Chapters/ValidateRemotely.html

        :param itunesiap.environment.Environment env: Override the environment.
        :param float timeout: The value is connection timeout of the verifying
            request. The default value is 30.0 when no `env` is given.
        :param bool use_production: The value is weather verifying in
            production server or not. The default value is :class:`bool` True
            when no `env` is given.
        :param bool use_sandbox: The value is weather verifying in
            sandbox server or not. The default value is :class:`bool` False
            when no `env` is given.

        :param bool verify_ssl: The value will be ignored.

        :return: :class:`itunesiap.receipt.Receipt` object if succeed.
        :raises: Otherwise raise a request exception.
        """
        env = options.get('env', default_env)
        use_production = options.get('use_production', env.use_production)
        use_sandbox = options.get('use_sandbox', env.use_sandbox)
        verify_ssl = options.get('verify_ssl', env.verify_ssl)  # noqa
        timeout = options.get('timeout', env.timeout)

        response = None
        if use_production:
            try:
                response = await self.aioverify_from(self.PRODUCTION_VALIDATION_URL, timeout=timeout)
            except exceptions.InvalidReceipt as e:
                if not use_sandbox or e.status != self.STATUS_SANDBOX_RECEIPT_ERROR:
                    raise
        if not response and use_sandbox:
            try:
                response = await self.aioverify_from(self.SANDBOX_VALIDATION_URL, timeout=timeout)
            except exceptions.InvalidReceipt:
                raise
        return response



""":mod:`itunesiap.receipt`

A successful response returns a JSON object including receipts. To manipulate
them in convinient way, `itunes-iap` wrapped it with :class:`ObjectMapper`.
"""
import datetime
import warnings
import pytz
import dateutil.parser
import json
from collections import defaultdict
from prettyexc import PrettyException

from .tools import lazy_property

__all__ = ('WARN_UNDOCUMENTED_FIELDS', 'Response', 'Receipt', 'InApp')


WARN_UNDOCUMENTED_FIELDS = True
WARN_UNLISTED_FIELDS = True

_warned_undocumented_fields = defaultdict(bool)
_warned_unlisted_field = defaultdict(bool)

'''
class ExpirationIntent(Enum):
    CustomerCanceledTheirSubscription = 1
    BillingError = 2
    CustumerDidNotAgreeToARecentPriceIncrease = 3
    ProductWasNotAvailableForPurchaseAtTheTimeOfRenewal = 4
    UnknownError = 5
'''


class MissingFieldError(PrettyException, AttributeError, KeyError):
    """A Backward compatibility error."""


def _rfc3339_to_datetime(value):
    """Try to parse Apple iTunes receipt date format.

    By reference, they insists it is rfc3339:

        https://developer.apple.com/library/ios/releasenotes/General/ValidateAppStoreReceipt/Chapters/ReceiptFields.html#//apple_ref/doc/uid/TP40010573-CH106-SW4

    Though data I got from apple server does not. So the strategy is:

        - Give them a chance anyway.
        - Or split timezone string and read it from pytz.
    """
    try:
        d = dateutil.parser.parse(value)
    except ValueError as e:
        value, timezone = value.rsplit(' ', 1)
        try:
            d = dateutil.parser.parse(value + '+00:00')
        except ValueError:
            raise e
        d = d.replace(tzinfo=pytz.timezone(timezone))
    return d


def _ms_to_datetime(value):
    nd = datetime.datetime.utcfromtimestamp(int(value) / 1000)
    ad = nd.replace(tzinfo=pytz.UTC)
    return ad


def _to_bool(data):
    assert data in ('true', 'false'), \
        ("Cannot convert {0}, "
         "acceptable values are true' and 'false'".format(data))
    return json.loads(data)


class ObjectMapper(object):
    """A pretty interface for decoded receipt object.

    `__DOCUMENTED_FIELDS__` and `__UNDOCUMENTED_FIELDS__` are managed lists of
    field names. They are regarded as safe values and guaranteed to be
    converted in python representation when needed.
    When a field exists in `__OPAQUE_FIELDS__`, its result will be redirected.
    The common type is :class:`str`.
    When a field exists in `__FIELD_ADAPTERS__`, it will be converted to
    corresponding python data representation.

    To access to the converted value, use a dictionary key as an attribute name.
    For example, the key `receipt` is accessible by:

    .. sourcecode:: python

        >>> mapper.receipt  # return converted python object Receipt
        >>> # == Receipt(mapper['receipt'])

    To access to the raw JSON value, use a dictionary key as an attribute name
    but with the prefix `_`. For example, the key `receipt` is accessible by:

        >>> mapper._receipt  # return converted python object Receipt
        >>> # == mapper['receipt']

    :param dict data: A JSON object.
    :return: :class:`ObjectMapper`
    """
    __OPAQUE_FIELDS__ = frozenset([])
    __FIELD_ADAPTERS__ = {}
    __DOCUMENTED_FIELDS__ = frozenset([])
    __UNDOCUMENTED_FIELDS__ = frozenset([])

    def __init__(self, data):
        self._ = data

    def __repr__(self):
        return u'<{self.__class__.__name__}({self._})>'.format(self=self)

    def __getitem__(self, item):
        return self._[item]

    def __contains__(self, item):
        return item in self._

    def __getattr__(self, name):
        try:
            return self.__getattribute__(name)
        except AttributeError:
            if name.startswith('_'):
                key = name[1:]
                if key in self.__DOCUMENTED_FIELDS__:
                    pass
                elif key in self.__UNDOCUMENTED_FIELDS__:
                    self.warn_undocumented(key)
                else:
                    self.warn_unlisted(key)
                try:
                    return self._[key]
                except KeyError:
                    raise MissingFieldError(name)

            if name in self.__DOCUMENTED_FIELDS__:
                pass
            elif name in self.__UNDOCUMENTED_FIELDS__:
                self.warn_undocumented(name)
            else:
                self.warn_unlisted(name)

            if name in self.__OPAQUE_FIELDS__:
                def _get(_self):
                    try:
                        value = _self._[name]
                    except KeyError:
                        raise MissingFieldError(name)
                    return value
                setattr(self.__class__, name, property(_get))
            elif name in self.__FIELD_ADAPTERS__:
                adapter = self.__FIELD_ADAPTERS__[name]
                if isinstance(adapter, tuple):
                    data_key, transform = adapter
                else:
                    data_key = name
                    transform = adapter

                def _get(_self):
                    try:
                        value = _self._[data_key]
                    except KeyError:
                        raise MissingFieldError(name)
                    return transform(value)
                setattr(self.__class__, name, lazy_property(_get))
            else:
                pass  # unhandled. raise AttributeError
        return self.__getattribute__(name)

    @classmethod
    def from_list(cls, data_list):
        return [cls(data) for data in data_list]

    @staticmethod
    def warn_undocumented(name):
        if not WARN_UNDOCUMENTED_FIELDS or _warned_undocumented_fields[name]:
            return
        warnings.warn(
            "The given field '{name}' is an undocumented field. "
            "The behavior is neither documented nor guaranteed by Apple. "
            "To suppress further warnings, set the "
            "`itunesiap.receipt.WARN_UNDOCUMENTED_FIELDS` False"
            .format(name=name))
        _warned_undocumented_fields[name] = True

    @staticmethod
    def warn_unlisted(name):
        if not WARN_UNLISTED_FIELDS or _warned_unlisted_field[name]:
            return
        warnings.warn(
            "The given field '{name}' is an unlisted field. "
            "If the field actually exists, ignore this warning message. "
            "To suppress further warnings, please report it to itunes-iap "
            "project or set the `itunesiap.receipt.WARN_UNLISTED_FIELDS` "
            "False.".format(name=name))
        _warned_unlisted_field[name] = True


class Receipt(ObjectMapper):
    """The actual receipt.

    The receipt may hold only one purchase directly in receipt object or may
    hold multiple purchases in `in_app` key.
    This object encapsulate it to list of :class:`InApp` object in `in_app`
    property.

    :see: `<https://developer.apple.com/library/content/releasenotes/General/ValidateAppStoreReceipt/Chapters/ReceiptFields.html>`_
    """
    __OPAQUE_FIELDS__ = frozenset([
        # app receipt fields
        'bundle_id',
        'application_version',
        'original_application_version',
        # in-app purchase receipt fields
        'product_id',
        'transaction_id',
        'original_transaction_id',
        'expires_date_formatted',
        'app_item_id',
        'version_external_identifier',
        'web_order_line_item_id',
        'auto_renew_product_id',
    ])
    __FIELD_ADAPTERS__ = {
        # app receipt fields
        'receipt_creation_date': _rfc3339_to_datetime,
        'receipt_creation_date_ms': int,
        'expiration_date': _rfc3339_to_datetime,
        'expiration_date_ms': int,
        # in-app purchase receipt fields
        'quantity': int,
        'purchase_date': _rfc3339_to_datetime,
        'purchase_date_ms': int,
        'original_purchase_date': _rfc3339_to_datetime,
        'original_purchase_date_ms': int,
        'expires_date': _ms_to_datetime,
        'expires_date_ms': ('expires_date', int),
        'expiration_intent': int,
        'is_in_billing_retry_period': _to_bool,
        'is_in_intro_offer_period': _to_bool,
        'cancellation_date': _rfc3339_to_datetime,
        'cancellation_reason': int,
        'auto_renew_status': int,
        'price_consent_status': int,
        'request_date': _rfc3339_to_datetime,
        'request_date_ms': int,
    }
    __DOCUMENTED_FIELDS__ = frozenset([
        # app receipt fields
        'bundle_id',
        'in_app',
        'application_version',
        'original_application_version',
        'receipt_creation_date',
        'expiration_date',
        # in-app purchase receipt fields
        'quantity',
        'product_id',
        'transaction_id',
        'original_transaction_id',
        'purchase_date',  # _formatted value
        'original_purchase_date',
        'expires_date',  # _ms value
        'is_in_billing_retry_period',
        'is_in_intro_offer_period',
        'cancellation_date',
        'cancellation_reason',
        'app_item_id',
        'version_external_identifier',
        'web_order_line_item_id',
        'auto_renew_status',
        'auto_renew_product_id',
        'price_consent_status',
    ])
    __UNDOCUMENTED_FIELDS__ = frozenset([
        # app receipt fields
        'request_date',
        'request_date_ms',
        'version_external_identifier',
        'receipt_creation_date_ms',
        'expiration_date_ms',
        # in-app purchase receipt fields
        'purchase_date_ms',
        'original_purchase_date_ms',
        'expires_date_formatted',
        'unique_identifier',
    ])

    @lazy_property
    def single_purchase(self):
        return Purchase(self._)

    @lazy_property
    def in_app(self):
        """The list of purchases. If the receipt has receipt keys in the
        receipt body, it still will be wrapped as an :class:`InApp` and consists
        of this property
        """
        if 'in_app' in self._:
            return list(map(InApp, self._in_app))
        else:
            return [self.single_purchase]

    @property
    def last_in_app(self):
        """The last item in `in_app` property order by purchase_date."""
        return sorted(
            self.in_app, key=lambda x: x['original_purchase_date_ms'])[-1]


class Purchase(ObjectMapper):
    """The individual purchases.

    The major keys are `quantity`, `product_id` and `transaction_id`.
    `quantty` and `product_id` mean what kind of product and
    and how many of them the customer bought. `unique_identifier` and
    `transaction_id` is used to check where it is processed and track related
    purchases.

    For the detail, see also Apple docs.

    Any `date` related keys will be converted to python
    :class:`datetime.datetime` object. The quantity and any `date_ms` related
    keys will be converted to python :class:`int`.
    """
    __OPAQUE_FIELDS__ = frozenset([
        'product_id',
        'transaction_id',
        'original_transaction_id',
        'web_order_line_item_id',
        'unique_identifier',
        'expires_date_formatted',
    ])
    __FIELD_ADAPTERS__ = {
        'quantity': int,
        'purchase_date': _rfc3339_to_datetime,
        'purchase_date_ms': int,
        'original_purchase_date': _rfc3339_to_datetime,
        'original_purchase_date_ms': int,
        'expires_date': _rfc3339_to_datetime,
        'expires_date_ms': int,
        'is_trial_period': _to_bool,
        'cancellation_date': _rfc3339_to_datetime,
        'cancellation_date_ms': int,
        'cancellation_reason': int,
    }
    __DOCUMENTED_FIELDS__ = frozenset([
        'quantity',
        'product_id',
        'transaction_id',
        'original_transaction_id',
        'purchase_date',
        'original_purchase_date',
        'expires_date',
        'is_trial_period',
        'cancellation_date',
        'cancellation_reason',
        'web_order_line_item_id',
    ])
    __UNDOCUMENTED_FIELDS__ = frozenset([
        'unique_identifier',
        'purchase_date_ms',
        'original_purchase_date_ms',
        'cancellation_date_ms',
        'expires_date_formatted',  # legacy receipts has this field as actual "expires_date"
    ])

    def __eq__(self, other):
        if not isinstance(other, Purchase):  # pragma: no cover
            return False
        return self._ == other._

    @lazy_property
    def expires_date(self):
        if 'expires_date_formatted' in self:
            return _rfc3339_to_datetime(self['expires_date_formatted'])
        try:
            value = self['expires_date']
        except KeyError:
            raise MissingFieldError('expires_date')
        try:
            int(value)
        except ValueError:
            return _rfc3339_to_datetime(value)
        else:
            return _ms_to_datetime(value)


class InApp(Purchase):
    pass


class PendingRenewalInfo(ObjectMapper):
    __OPAQUE_FIELDS__ = frozenset([
        'auto_renew_product_id',
    ])
    __FIELD_ADAPTERS__ = {
        'auto_renew_status': int,
        'expiration_intent': int,
        'is_in_billing_retry_period': int,
    }
    __DOCUMENTED_FIELDS__ = frozenset([
        'expiration_intent',
        'auto_renew_status',
        'auto_renew_product_id',
        'is_in_billing_retry_period',
    ])
    __UNDOCUMENTED_FIELDS__ = frozenset([
    ])


class Response(ObjectMapper):
    """The root response.

    About the value of status:
        - See https://developer.apple.com/library/ios/releasenotes/General/ValidateAppStoreReceipt/Chapters/ValidateRemotely.html#//apple_ref/doc/uid/TP40010573-CH104-SW1
    """
    __OPAQUE_FIELDS__ = frozenset([
        'latest_receipt',
    ])
    __FIELD_ADAPTERS__ = {
        'status': int,
        'receipt': Receipt,
        'pending_renewal_info': PendingRenewalInfo.from_list,
    }
    __DOCUMENTED_FIELDS__ = frozenset([
        'status',
        'receipt',
        'latest_receipt',
        'latest_receipt_info',
        # 'latest_expired_receipt_info',
        'pending_renewal_info',
        # is-retryable
    ])
    __UNDOCUMENTED_FIELDS__ = frozenset([
    ])

    @lazy_property
    def latest_receipt_info(self):
        if 'latest_receipt_info' not in self:
            # not an auto-renew purchase
            raise MissingFieldError('latest_receipt_info')
        info = self['latest_receipt_info']
        if isinstance(info, dict):  # iOS6 style
            return Purchase(info)
        elif isinstance(info, list):  # iOS7 style
            return InApp.from_list(info)
        else:  # pragma: no cover
            assert False
